//
// Created by mmath on 12/3/18.
//

#ifndef PYSCAN_CONVEXHULL_HPP
#define PYSCAN_CONVEXHULL_HPP

#include "Point.hpp"

namespace pyscan {

    point_list_t graham_march(point_list_t points);
}
#endif //PYSCAN_CONVEXHULL_HPP
