/***************
 * header.h
 * important declarations for all data structures.
 *
 * Jeff Phillips
 * 5.21.02
 **************/

#ifndef HEADER__H
#define HEADER__H

#include "stdlib.h"
#include "stdio.h"

#define TRUE 1
#define FALSE 0

#ifndef NULL
#define NULL 0
#endif

#ifndef PI
#define PI 3.14159265358979323846 
#endif

#define RED 1
#define BLUE 0
#define BASE 2

#endif
